@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2006-2007  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -48,6 +48,7 @@ struct target;
 struct sink;
 struct headset;
 struct gateway;
+struct dev_priv;
 
 struct audio_device {
 	struct btd_device *btd_dev;
@@ -57,18 +58,24 @@ struct audio_device {
 	bdaddr_t src;
 	bdaddr_t dst;
 
+	gboolean auto_connect;
+
 	struct headset *headset;
 	struct gateway *gateway;
 	struct sink *sink;
 	struct source *source;
 	struct control *control;
 	struct target *target;
+
+	struct dev_priv *priv;
 };
 
-struct audio_device *device_register(DBusConnection *conn,
+struct audio_device *audio_device_register(DBusConnection *conn,
+					struct btd_device *device,
 					const char *path, const bdaddr_t *src,
 					const bdaddr_t *dst);
 
-void device_unregister(struct audio_device *device);
+void audio_device_unregister(struct audio_device *device);
 
-gboolean device_is_connected(struct audio_device *dev, const char *interface);
+gboolean audio_device_is_connected(struct audio_device *dev,
+						const char *interface);
