@@ -64,10 +64,18 @@
 #define DISCOVERY_TIMER		2
 
 struct btd_driver_data {
+	guint id;
 	struct btd_device_driver *driver;
 	void *priv;
 };
 
+struct btd_disconnect_data {
+	guint id;
+	disconnect_watch watch;
+	void *user_data;
+	GDestroyNotify destroy;
+};
+
 struct bonding_req {
 	DBusConnection *conn;
 	DBusMessage *msg;
@@ -91,6 +99,7 @@ struct btd_device {
 	struct btd_adapter	*adapter;
 	GSList		*uuids;
 	GSList		*drivers;		/* List of driver_data */
+	GSList		*watches;		/* List of disconnect_data */
 	gboolean	temporary;
 	struct agent	*agent;
 	guint		disconn_timer;
@@ -116,6 +125,7 @@ struct btd_device {
 	gboolean	renewed_key;
 
 	gboolean	authorizing;
+	gint		ref;
 };
 
 struct browse_req {
@@ -181,7 +191,8 @@ static void device_free(gpointer user_data)
 	if (device->agent)
 		agent_destroy(device->agent, FALSE);
 
-	if (agent && agent_is_busy(agent, device))
+	if (agent && (agent_is_busy(agent, device) ||
+				agent_is_busy(agent, device->authr)))
 		agent_cancel(agent);
 
 	g_slist_foreach(device->uuids, (GFunc) g_free, NULL);
@@ -193,6 +204,9 @@ static void device_free(gpointer user_data)
 	if (device->discov_timer)
 		g_source_remove(device->discov_timer);
 
+	debug("device_free(%p)", device);
+
+	g_free(device->authr);
 	g_free(device->path);
 	g_free(device);
 }
@@ -412,9 +426,35 @@ static DBusMessage *set_property(DBusConnection *conn,
 	return invalid_args(msg);
 }
 
+static void browse_req_free(struct browse_req *req)
+{
+	struct btd_device *device = req->device;
+
+	device->discov_active = 0;
+
+	if (device->discov_requestor) {
+		g_dbus_remove_watch(req->conn, device->discov_listener);
+		device->discov_listener = 0;
+		g_free(device->discov_requestor);
+		device->discov_requestor = NULL;
+	}
+
+	if (req->msg)
+		dbus_message_unref(req->msg);
+	if (req->conn)
+		dbus_connection_unref(req->conn);
+	g_slist_foreach(req->profiles_added, (GFunc) g_free, NULL);
+	g_slist_free(req->profiles_added);
+	g_slist_free(req->profiles_removed);
+	if (req->records)
+		sdp_list_free(req->records, (sdp_free_func_t) sdp_record_free);
+	g_free(req);
+}
+
 static void discover_services_req_exit(DBusConnection *conn, void *user_data)
 {
-	struct btd_device *device = user_data;
+	struct browse_req *req = user_data;
+	struct btd_device *device = req->device;
 	struct btd_adapter *adapter = device->adapter;
 	bdaddr_t src;
 
@@ -423,6 +463,8 @@ static void discover_services_req_exit(DBusConnection *conn, void *user_data)
 	debug("DiscoverDevices requestor exited");
 
 	bt_cancel_discovery(&src, &device->bdaddr);
+
+	browse_req_free(req);
 }
 
 static DBusMessage *discover_services(DBusConnection *conn,
@@ -520,6 +562,30 @@ fail:
 	return FALSE;
 }
 
+void device_disconnect(struct btd_device *device)
+{
+	GSList *l;
+
+	l = device->watches;
+	while (l) {
+		struct btd_disconnect_data *data = l->data;
+
+		l = l->next;
+
+		if (data->watch)
+			/* temporary is set if device is going to be removed */
+			data->watch(device, device->temporary,
+					data->user_data);
+	}
+
+	g_slist_foreach(device->watches, (GFunc) g_free, NULL);
+	g_slist_free(device->watches);
+	device->watches = NULL;
+
+	device->disconn_timer = g_timeout_add_seconds(DISCONNECT_TIMER,
+						disconnect_timeout, device);
+}
+
 static DBusMessage *disconnect(DBusConnection *conn, DBusMessage *msg,
 							void *user_data)
 {
@@ -534,8 +600,7 @@ static DBusMessage *disconnect(DBusConnection *conn, DBusMessage *msg,
 			DEVICE_INTERFACE, "DisconnectRequested",
 			DBUS_TYPE_INVALID);
 
-	device->disconn_timer = g_timeout_add_seconds(DISCONNECT_TIMER,
-						disconnect_timeout, device);
+	device_disconnect(device);
 
 	return dbus_message_new_method_return(msg);
 }
@@ -613,6 +678,42 @@ gboolean device_has_connection(struct btd_device *device, uint16_t handle)
 	return (handle == device->handle);
 }
 
+guint device_add_disconnect_watch(struct btd_device *device,
+				disconnect_watch watch, void *user_data,
+				GDestroyNotify destroy)
+{
+	struct btd_disconnect_data *data;
+	static guint id = 0;
+
+	data = g_new0(struct btd_disconnect_data, 1);
+	data->id = ++id;
+	data->watch = watch;
+	data->user_data = user_data;
+	data->destroy = destroy;
+
+	device->watches = g_slist_append(device->watches, data);
+
+	return data->id;
+}
+
+void device_remove_disconnect_watch(struct btd_device *device, guint id)
+{
+	GSList *l;
+
+	for (l = device->watches; l; l = l->next) {
+		struct btd_disconnect_data *data = l->data;
+
+		if (data->id == id) {
+			device->watches = g_slist_remove(device->watches,
+							data);
+			if (data->destroy)
+				data->destroy(data->user_data);
+			g_free(data);
+			return;
+		}
+	}
+}
+
 void device_set_secmode3_conn(struct btd_device *device, gboolean enable)
 {
 	device->secmode3 = enable;
@@ -654,7 +755,7 @@ struct btd_device *device_create(DBusConnection *conn,
 
 	device->auth = 0xff;
 
-	return device;
+	return btd_device_ref(device);
 }
 
 void device_set_name(struct btd_device *device, const char *name)
@@ -698,12 +799,6 @@ static void device_remove_bonding(struct btd_device *device,
 	ba2str(&bdaddr, srcaddr);
 	ba2str(&device->bdaddr, dstaddr);
 
-	dev_id = adapter_get_dev_id(device->adapter);
-
-	dd = hci_open_dev(dev_id);
-	if (dd < 0)
-		return;
-
 	create_name(filename, PATH_MAX, STORAGEDIR, srcaddr,
 			"linkkeys");
 
@@ -715,6 +810,12 @@ static void device_remove_bonding(struct btd_device *device,
 	if (!paired)
 		return;
 
+	dev_id = adapter_get_dev_id(device->adapter);
+
+	dd = hci_open_dev(dev_id);
+	if (dd < 0)
+		return;
+
 	/* Delete the link key from storage */
 	textfile_casedel(filename, dstaddr);
 
@@ -755,14 +856,13 @@ void device_remove(struct btd_device *device, DBusConnection *conn,
 {
 	GSList *list;
 	struct btd_device_driver *driver;
-	gchar *path = g_strdup(device->path);
 
-	debug("Removing device %s", path);
+	debug("Removing device %s", device->path);
 
 	if (device->bonding)
 		device_cancel_bonding(device, HCI_OE_USER_ENDED_CONNECTION);
 
-	if (!device->temporary && remove_stored)
+	if (remove_stored)
 		device_remove_stored(device, conn);
 
 	for (list = device->drivers; list; list = list->next) {
@@ -773,10 +873,7 @@ void device_remove(struct btd_device *device, DBusConnection *conn,
 		g_free(driver_data);
 	}
 
-
-	g_dbus_unregister_interface(conn, path, DEVICE_INTERFACE);
-
-	g_free(path);
+	btd_device_unref(device);
 }
 
 gint device_address_cmp(struct btd_device *device, const gchar *address)
@@ -1169,31 +1266,6 @@ static void store_profiles(struct btd_device *device)
 	g_free(str);
 }
 
-static void browse_req_free(struct browse_req *req)
-{
-	struct btd_device *device = req->device;
-
-	device->discov_active = 0;
-
-	if (device->discov_requestor) {
-		g_dbus_remove_watch(req->conn, device->discov_listener);
-		device->discov_listener = 0;
-		g_free(device->discov_requestor);
-		device->discov_requestor = NULL;
-	}
-
-	if (req->msg)
-		dbus_message_unref(req->msg);
-	if (req->conn)
-		dbus_connection_unref(req->conn);
-	g_slist_foreach(req->profiles_added, (GFunc) g_free, NULL);
-	g_slist_free(req->profiles_added);
-	g_slist_free(req->profiles_removed);
-	if (req->records)
-		sdp_list_free(req->records, (sdp_free_func_t) sdp_record_free);
-	g_free(req);
-}
-
 static void search_cb(sdp_list_t *recs, int err, gpointer user_data)
 {
 	struct browse_req *req = user_data;
@@ -1254,6 +1326,8 @@ proceed:
 
 	g_dbus_send_message(req->conn, reply);
 
+	device_set_temporary(device, FALSE);
+
 cleanup:
 	browse_req_free(req);
 }
@@ -1352,7 +1426,7 @@ int device_browse(struct btd_device *device, DBusConnection *conn,
 		device->discov_listener = g_dbus_add_disconnect_watch(conn,
 						sender,
 						discover_services_req_exit,
-						device, NULL);
+						req, NULL);
 	}
 
 	err = bt_search_service(&src, &device->bdaddr,
@@ -1509,7 +1583,17 @@ DBusMessage *new_authentication_return(DBusMessage *msg, uint8_t status)
 	}
 }
 
-static void bonding_request_free(struct bonding_req *bonding, gboolean close)
+static void bonding_request_cancel(struct bonding_req *bonding)
+{
+	if (!bonding->io)
+		return;
+
+	g_io_channel_shutdown(bonding->io, TRUE, NULL);
+	g_io_channel_unref(bonding->io);
+	bonding->io = NULL;
+}
+
+static void bonding_request_free(struct bonding_req *bonding)
 {
 	struct btd_device *device;
 
@@ -1528,11 +1612,8 @@ static void bonding_request_free(struct bonding_req *bonding, gboolean close)
 	if (bonding->io_id)
 		g_source_remove(bonding->io_id);
 
-	if (bonding->io) {
-		if (close)
-			g_io_channel_shutdown(bonding->io, TRUE, NULL);
+	if (bonding->io)
 		g_io_channel_unref(bonding->io);
-	}
 
 	device = bonding->device;
 	g_free(bonding);
@@ -1616,6 +1697,8 @@ static gboolean bonding_io_cb(GIOChannel *io, GIOCondition cond,
 	error_connection_attempt_failed(device->bonding->conn,
 					device->bonding->msg, ENETDOWN);
 
+	bonding_request_free(device->bonding);
+
 	return FALSE;
 }
 
@@ -1705,7 +1788,7 @@ failed:
 
 cleanup:
 	device->bonding->io_id = 0;
-	bonding_request_free(device->bonding, FALSE);
+	bonding_request_free(device->bonding);
 }
 
 static void create_bond_req_exit(DBusConnection *conn, void *user_data)
@@ -1719,7 +1802,7 @@ static void create_bond_req_exit(DBusConnection *conn, void *user_data)
 
 	if (device->bonding) {
 		device->bonding->listener_id = 0;
-		bonding_request_free(device->bonding, TRUE);
+		bonding_request_cancel(device->bonding);
 	}
 }
 
@@ -1813,7 +1896,6 @@ void device_bonding_complete(struct btd_device *device, uint8_t status)
 		goto failed;
 
 	device->auth = 0xff;
-	device->temporary = FALSE;
 
 	g_free(device->authr);
 	device->authr = NULL;
@@ -1847,7 +1929,7 @@ void device_bonding_complete(struct btd_device *device, uint8_t status)
 
 	device_set_paired(device, TRUE);
 
-	bonding_request_free(bonding, TRUE);
+	bonding_request_free(bonding);
 
 	return;
 
@@ -1884,7 +1966,8 @@ void device_cancel_bonding(struct btd_device *device, uint8_t status)
 	reply = new_authentication_return(bonding->msg, status);
 	g_dbus_send_message(bonding->conn, reply);
 
-	bonding_request_free(bonding, TRUE);
+	bonding_request_cancel(bonding);
+	bonding_request_free(bonding);
 }
 
 static void pincode_cb(struct agent *agent, DBusError *err, const char *pincode,
@@ -2109,3 +2192,31 @@ void btd_unregister_device_driver(struct btd_device_driver *driver)
 {
 	device_drivers = g_slist_remove(device_drivers, driver);
 }
+
+struct btd_device *btd_device_ref(struct btd_device *device)
+{
+	device->ref++;
+
+	debug("btd_device_ref(%p): ref=%d", device, device->ref);
+
+	return device;
+}
+
+void btd_device_unref(struct btd_device *device)
+{
+	DBusConnection *conn = get_dbus_connection();
+	gchar *path;
+
+	device->ref--;
+
+	debug("btd_device_unref(%p): ref=%d", device, device->ref);
+
+	if (device->ref > 0)
+		return;
+
+	path = g_strdup(device->path);
+
+	g_dbus_unregister_interface(conn, path, DEVICE_INTERFACE);
+
+	g_free(path);
+}
